program unblur

    use Globals
    use UserSuppliedParameters
    use UserInputs
    use ImageFiles
    use Images
    use Peaks
    use Curves
    use ProgressBars
    use NumericTextFiles
    use UsefulFunctions
    use StringManipulations
    use ElectronDoses
    use ContrastTransferFunctions
    use unblur_functions, only : unblur_refine_alignment
    !use ctffind_routines, only : ParseCTFFindResultsFile

    implicit none

    ! Variables associated with user input

    type(UserInput)                 ::  my_user_input

    type(UserSuppliedInteger)       ::  number_of_frames_per_movie

    type(UserSuppliedInteger)       ::  maximum_number_of_iterations


    type(UserSuppliedInteger)       ::  width_of_vertical_line
    type(UserSuppliedInteger)       ::  width_of_horizontal_line

    type(UserSuppliedReal)          ::  pixel_size
    type(UserSuppliedReal)          ::  inner_radius
    type(UserSuppliedReal)          ::  outer_radius
    type(UserSuppliedReal)          ::  bfactor
    type(UserSuppliedReal)          ::  threshold_maximum_shift

    type(UserSuppliedFilename)      ::  input_filename
    type(UserSuppliedFilename)      ::  output_filename
    type(UserSuppliedFilename)      ::  aligned_frames_filename
    type(UserSuppliedFilename)      ::  shifts_output_filename
    type(UserSuppliedFilename)      ::  frc_output_filename

    type(UserSuppliedLogical)       ::  apply_drift_filter
    type(UserSuppliedLogical)       ::  mask_central_cross
    type(UserSuppliedLogical)       ::  set_expert_options
    type(UserSuppliedLogical)       ::  verbose_output

    type(UserSuppliedLogical)       ::  apply_dose_filter
    type(UserSuppliedLogical)       ::  divide_dose_filter
    type(UserSuppliedLogical)       ::  save_aligned_frames
    type(UserSuppliedReal)          ::  acceleration_voltage
    type(UserSuppliedReal)          ::  exposure_per_frame
    type(UserSuppliedReal)          ::  pre_exposure_amount

    !type(UserSuppliedLogical)       ::  apply_ctf
    !type(UserSuppliedLogical)       ::  phase_flip_only
    !type(UserSuppliedFilename)      ::  ctffind_results

    ! Variable associated with Images

    type(ImageFile)                 ::  input_imagefile

    type(Image),    allocatable     ::  image_stack(:)
    type(Image),    allocatable     ::  unbinned_image_stack(:)
    type(Image)                     ::  sum_image
    type(Image)                     ::  sum_image_2


    real                            ::  maximum_shift
    real                            ::  maximum_x_shift
    real                            ::  maximum_y_shift
    real                            ::  current_drift_x
    real                            ::  current_drift_y

    real                            ::  final_score
    real                            ::  final_score_numerator
    real                            ::  final_score_denominator

    real                            ::  average_x_shift
    real                            ::  average_y_shift

    integer                         ::  number_of_input_images
    integer                         ::  number_of_movies

    integer                         ::  image_counter
    integer                         ::  bin_counter
    integer                         ::  movie_counter

    integer                         ::  iterations_performed

    integer                         ::  order_of_polynomial
    integer                         ::  number_of_middle_image


    type(Curve)                     ::  x_shifts
    type(Curve)                     ::  y_shifts
    type(Curve)                     ::  fsc_curve

    type(ProgressBar)               ::  my_progress_bar

    type(NumericTextFile)           ::  my_shifts_file
    type(NumericTextFile)           ::  my_frc_file

    real, allocatable               ::  current_x_shifts(:)
    real, allocatable               ::  current_y_shifts(:)
    real, allocatable               ::  old_x_shifts(:)
    real, allocatable               ::  old_y_shifts(:)
    real, allocatable               ::  total_x_shifts(:)
    real, allocatable               ::  total_y_shifts(:)

    real, allocatable               ::  frc_x_data(:)
    real, allocatable               ::  frc_y_data(:)

    !real                            ::  temp_real(2) ! for writing shifts to file

    ! Option for pre-binning
    logical,        parameter       ::  allow_pre_binning = .true.
    integer                         ::  pre_binning_factor
    integer                         ::  binned_dimensions(3)
    real                            ::  binned_pixel_size
    real                            ::  inner_radius_after_binning
    real                            ::  outer_radius_after_binning

    ! Dose filtering
    type(ElectronDose)              ::  my_electron_dose
    !type(Image),    allocatable     ::  dose_filters_squared(:)
    !type(Image)                     ::  dose_filters_squared_sum

    integer                         ::  i,j
    real                            ::  x,y, x_sq, y_sq
    real                            ::  azimuth
    real                            ::  current_denominator
    real                            ::  current_critical_dose
    real                            ::  current_optimal_dose
    real                            ::  current_dose
    real                            ::  current_ctf
    real,           parameter       ::  minimum_absolute_ctf_value = 0.001

    type(ContrastTransferFunction), allocatable ::  frame_ctf(:)



    ! Initialise this_program with the program name and version number
    call this_program%Init('UnBlur', '1.0.2','2015')


    ! Do the user input
    call my_user_input%Init(this_program%program_name)

    input_filename   = my_user_input%GetFilenameFromUser('Input stack filename', &
                                   'The input file, containing your raw movie frames', 'input_filename', 'my_movie.mrc', .true.)

    ! Open the input file
    call input_imagefile%Init(input_filename%value)
    number_of_input_images = input_imagefile%GetStackSize()

    number_of_frames_per_movie      =   my_user_input%GetIntegerFromUser('Number of frames per movie', &
                                        'How many frames per micrograph?', &
                                        'number_of_frames_per_movie',IntegerToString(number_of_input_images), &
                                        min_value=1,max_value=number_of_input_images)

    output_filename                 =   my_user_input%GetFilenameFromUser('Output aligned sum file', &
                                        'The output file, containing a weighted sum of the aligned input frames', &
                                        'output_filename', 'my_aligned_sum.mrc', .false.)

    shifts_output_filename          =   my_user_input%GetFilenameFromUser('Output shifts file', &
                                        'This file will contain the computed X/Y shifts for each frame.', &
                                        'shifts_filename', 'my_shifts.txt', .false.)

    pixel_size                      =   my_user_input%GetRealFromUser('Pixel size of images (A)', &
                                        'Pixel size of input images in Angstroms', 'pixel_size', '1', min_value = 0.0e0)

    apply_dose_filter               =   my_user_input%GetLogicalFromUser('Apply Dose filter?', &
                                        'Apply a dose-dependent filter to frames before summing them', &
                                        'apply_dose_filter','NO')

    if (apply_dose_filter%value) then
         exposure_per_frame         =   my_user_input%GetRealFromUser('Exposure per frame (e/A^2)', &
                                        'Exposure per frame, in electrons per square Angstrom','exposure_per_frame', &
                                        '1.0',min_value=0.00001e0)
         acceleration_voltage       =   my_user_input%GetRealFromUser('Acceleration voltage (kV)', &
                                        'Acceleration voltage during imaging','acceleration_voltage','300.0', &
                                        min_value=200.0e0,max_value=300.0e0)

         pre_exposure_amount         =   my_user_input%GetRealFromUser('Pre-exposure amount(e/A^2)', &
                                        'Amount of pre-exposure prior to the first frame, in electrons per square Angstrom','pre_exposure_amount', &
                                        '0.0',min_value=0.0000)

    endif

    save_aligned_frames             =   my_user_input%GetLogicalFromUser('Save Aligned Frames?', &
                                        'If yes, the aligned frames will be saved to the supplied filename', &
                                        'save_aligned_frames','NO')

    if (save_aligned_frames%value) then

    aligned_frames_filename          =   my_user_input%GetFilenameFromUser('Aligned frames output filename', &
                                        'The output file, containing the aligned frames', &
                                        'aligned_frames_filename', 'my_aligned_frames.mrc', .false.)


    endif



    set_expert_options              =   my_user_input%GetLogicalFromUser('Set Expert Options?', &
                                        'Set these for more control, hopefully not needed', 'set_expert_options', 'NO')



    if (set_expert_options%value) then

       ! apply_ctf                       =   my_user_input%GetLogicalFromUser('Apply CTF to frames?', &
       !                                     'Apply CTF to frames before aligning them','apply_ctf','no')

       ! if (apply_ctf%value) then
       !     phase_flip_only             =   my_user_input%GetLogicalFromUser('Phase flip only?', &
       !                                     'Multiply by the sign of the CTF rather than its value','phase_flip_only','no')

       !     ctffind_results             =   my_user_input%GetFilenameFromUser('Results from CTFfind', &
       !                                     'Filename of CTFfind output','ctffind_results','my_ctf_results.txt', &
       !                                     file_must_exist=.true.)
       ! endif

        frc_output_filename         =   my_user_input%GetFilenameFromUser('Output FRC file', &
                                        'This file will contain the computed FSC of the two half sums', &
                                        'frc_filename', 'my_frc.txt', .false.)


        inner_radius                =   my_user_input%GetRealFromUser('Minimum shift for initial search (Angstroms)', &
                                        'Initial search will be limited to between the inner and outer radii.  ', &
                                        'inner_radius', '2.0', min_value = 0.0e0)

        outer_radius                =   my_user_input%GetRealFromUser('Outer radius shift limit (Angstroms)', &
                                        'The maximum shift of each alignment step will be limited to this value. ', &
                                        'outer_radius', '200.0', min_value = inner_radius%value)

        bfactor                     =   my_user_input%GetRealFromUser('B-factor to apply to images (A^2)', &
                                        'Initial search will be limited to between the inner and outer radii', &
                                        'bfactor', '1500',min_value=0.0)

        !apply_drift_filter          =   my_user_input%GetLogicalFromUser('Apply Drift Filter?', &
        !                                'If yes, a corrective drift filter will be  applied', 'use_drift_filter', 'NO')
	apply_drift_filter%value = .false.


        !mask_central_cross          =   my_user_input%GetLogicalFromUser('Mask central cross in Fourier space?', &
        !                                'This can help reduce problems caused by line artifacts from the detector', &
        !                                'mask_with_cross', 'YES')
        mask_central_cross%value    =   .true.
        mask_central_cross%set_by_user = .false.

        if (mask_central_cross%value) then

            width_of_vertical_line  =   my_user_input%GetIntegerFromUser('Half-width of central vertical line of Fourier mask', &
                                        'The vertical line mask will be twice this size. The central cross mask helps'//&
                                        ' reduce problems by line artefacts from the detector', 'vertical_width', '1', 1)

            width_of_horizontal_line=   my_user_input%GetIntegerFromUser('Half-width of central horizontal line of Fourier mask', &
                                        'The horizontal line mask will be twice this size. The central cross mask helps'//&
                                        ' reduce problems by line artefacts from the detector', 'horizontal_width', '1', 1)
        endif

        threshold_maximum_shift     =   my_user_input%GetRealFromUser('Termination shift threshold', &
                                        'Alignment will iterate until the maximum shift is below this value', &
                                        'threshold_shift', '0.1')

        maximum_number_of_iterations=   my_user_input%GetIntegerFromUser('Maximum number of iterations', &
                                        'Alignment will stop at this number, even if the threshold shift is not reached', &
                                        'max_iterations', '10', min_value = 1)

        if (apply_dose_filter%value) then

            divide_dose_filter      =   my_user_input%GetLogicalFromUser('Restore Noise Power?', &
                                        'divide by dose filter squared', &
                                        'RESTORE_NOISE_POWER','YES')
        endif

        verbose_output              =   my_user_input%GetLogicalFromUser('Verbose Output?', &
                                        'All the output!', 'verbose_output', 'NO')
    else

        inner_radius%value = pixel_size%value * 2e0
        outer_radius%value = 80e0
        bfactor%value = 1500
        apply_drift_filter%value = .false.
        mask_central_cross%value = .true.
        width_of_vertical_line%value = 1e0
        width_of_horizontal_line%value = 1e0
        threshold_maximum_shift%value = pixel_size%value * 0.5
        maximum_number_of_iterations%value = 20
        verbose_output%value = .false.
        divide_dose_filter%value = .true.
        !apply_ctf%value = .false.
    endif

    call my_user_input%UpdateDefaults()
    write(*,*)

    !
    ! End of user interaction
    !


    ! Open the output file for the shifts
    call my_shifts_file%Init(shifts_output_filename%value, OPEN_TO_WRITE, number_of_frames_per_movie%value)


    ! Do some basic checks..
    number_of_movies = number_of_input_images / number_of_frames_per_movie%value

    if (number_of_frames_per_movie%value * number_of_movies .ne. number_of_input_images) then
        call this_program%TerminateWithFatalError('Unblur','Number of images in input stack not '//&
                                                            'divisible by number of frames per movie')
    endif

    if (number_of_frames_per_movie%value .le. 2) then
        call this_program%TerminateWithFatalError('Unblur','Need at least 3 images')
    endif

    ! which is the middle image

    if (IsOdd(number_of_frames_per_movie%value)) then
        number_of_middle_image = (number_of_frames_per_movie%value + 1) / 2
    else
        number_of_middle_image = number_of_frames_per_movie%value / 2 + 1
    endif

    ! Calculate order of polynomial to use..

    if (number_of_frames_per_movie%value .gt. 6) then
        order_of_polynomial = 6
    else
        order_of_polynomial = number_of_frames_per_movie%value - 1
    endif

    if (apply_dose_filter%value) then
        call my_electron_dose%Init(acceleration_voltage%value)

        !if (divide_dose_filter%value) then
        !    allocate(dose_filters_squared(number_of_frames_per_movie%value))
        !endif
    endif


    ! Allocation

    allocate(image_stack(number_of_frames_per_movie%value))
    allocate(unbinned_image_stack(number_of_frames_per_movie%value))
    allocate(current_x_shifts(number_of_frames_per_movie%value))
    allocate(current_y_shifts(number_of_frames_per_movie%value))
    allocate(total_x_shifts(number_of_frames_per_movie%value))
    allocate(total_y_shifts(number_of_frames_per_movie%value))
    allocate(old_x_shifts(number_of_frames_per_movie%value))
    allocate(old_y_shifts(number_of_frames_per_movie%value))

    ! Initialise the curves

    call x_shifts%Init(number_of_frames_per_movie%value)
    call y_shifts%Init(number_of_frames_per_movie%value)

    ! Parse CTFfind results
    !if (apply_ctf%value) then
    !    call ParseCTFFindResultsFile(ctffind_results%value,pixel_size%value,frame_ctf)
    !    if (size(frame_ctf) .ne. number_of_frames_per_movie%value*number_of_movies) then
    !        call this_program%TerminateWithFatalError('Unblur',&
    !                                'Number of CTF parameters = '//IntegerToString(size(frame_ctf))// &
    !                                'Number of frames = '//IntegerToString(number_of_frames_per_movie%value))
    !    endif
    !endif

    ! Loop over movies

    if (.not. verbose_output%value) then
        Write(*, '(a)') 'Performing Alignment...'
        if (number_of_movies .eq. 1) write(*,*)

        call my_progress_bar%Begin(number_of_movies)
    endif

    do movie_counter=1,number_of_movies

        !
        if (verbose_output%value) then
            write(*,'(//2(a,i0),a//)') 'Unblurring micrograph ', movie_counter, ' of ', number_of_movies, '...'

            ! Read the images
            write(*,'(a)') 'Reading images...'
            call my_progress_bar%Begin(number_of_frames_per_movie%value)
        endif

        do image_counter=1,number_of_frames_per_movie%value
            call image_stack(image_counter)%ReadFromDisk(input_imagefile, &
                                                         (movie_counter-1)*number_of_frames_per_movie%value+image_counter)

            if (verbose_output%value) call my_progress_bar%Update(image_counter)
        enddo

        if (verbose_output%value) call my_progress_bar%Finish()


        ! FFT
        if (verbose_output%value) then
            write(*,'(a)') 'Fourier transforming images...'
            call my_progress_bar%Begin(number_of_frames_per_movie%value)
        endif
        iterations_performed = 0
        !$omp parallel default(shared) private(image_counter)
        !$omp do
        do image_counter = 1,number_of_frames_per_movie%value

            call image_stack(image_counter)%ForwardFFT()

!            if (apply_ctf%value) then
!                call image_stack(image_counter)%ApplyCTF(frame_ctf( (movie_counter-1)*number_of_frames_per_movie%value &
!                                                                   +image_counter), &
!                                                         phase_flip_only%value, &
!                                                         minimum_absolute_ctf_value)
!            endif

            !$omp atomic
            iterations_performed = iterations_performed + 1
            if (verbose_output%value) call my_progress_bar%Update(iterations_performed)
        enddo
        !$omp enddo
        !$omp end parallel
        if (verbose_output%value) call my_progress_bar%Finish()


        ! If we're going to bin, let's find a good binning factor
        if (movie_counter .eq. 1) then
            if (allow_pre_binning) then
                ! We need a binning factor which leads to integer binned dimensions
                pre_binning_factor = nint(image_stack(1)%GetLogicalDimension(1) / 1024.0)

            else
                pre_binning_factor = 1
            endif
        endif

        ! Bin the images
        if (pre_binning_factor .gt. 1) then

            if (verbose_output%value) then
                write(*,'(a,i0,a)') 'Cropping Fourier transforms of images by a factor of ', &
                                    pre_binning_factor,' to speed-up frame alignment...'
            endif

            binned_dimensions = image_stack(1)%GetLogicalDimensions() / pre_binning_factor
            binned_dimensions(3) = 1
            binned_pixel_size = pixel_size%value * real(pre_binning_factor)
            !
            if (verbose_output%value) call my_progress_bar%Begin(number_of_frames_per_movie%value)
            iterations_performed = 0
            !$omp parallel default(shared) private(image_counter)
            !$omp do
            do image_counter = 1,number_of_frames_per_movie%value
                unbinned_image_stack(image_counter) = image_stack(image_counter)
                call image_stack(image_counter)%Resize(binned_dimensions)
                !$omp atomic
                iterations_performed = iterations_performed + 1
                if (verbose_output%value) call my_progress_bar%Update(iterations_performed)
            enddo

            !$omp enddo
            !$omp end parallel
            if (verbose_output%value) call my_progress_bar%Finish()
        else
            binned_dimensions = image_stack(1)%GetLogicalDimensions()
            binned_pixel_size = pixel_size%value
        endif



        ! Perform the initial alignment...
        if (verbose_output%value) write(*,'(a)') 'Performing initial alignment (ignoring central peak of CCF)...'

        inner_radius_after_binning = inner_radius%value / binned_pixel_size
        outer_radius_after_binning = outer_radius%value / binned_pixel_size

        if (inner_radius%value .ge. 1.0e0) then
            inner_radius_after_binning = max(inner_radius_after_binning,1.0e0)
        endif

        !write(*,'(a,f0.2)') '**debug(Unblur): shift threshold = ', threshold_maximum_shift%value / binned_pixel_size * pre_binning_factor
        !write(*,'(a,f0.2)') '**debug(Unblur): inner radius after binning = ', inner_radius_after_binning

        call unblur_refine_alignment(   image_stack, &
                                        1, & ! one iteration only
                                        bfactor%value / binned_pixel_size**2, &
                                        mask_central_cross%value, &
                                        width_of_vertical_line%value, &
                                        width_of_horizontal_line%value, &
                                        .false.,0.0e0, & ! no real-space masking of images
                                        inner_radius_after_binning, &
                                        outer_radius_after_binning, &
                                        max(threshold_maximum_shift%value/binned_pixel_size*pre_binning_factor,&
                                            inner_radius_after_binning+0.01), &
                                        current_x_shifts,current_y_shifts,binned_pixel_size)

        maximum_x_shift = maxval(abs(current_x_shifts))
        maximum_y_shift = maxval(abs(current_y_shifts))

        if (maximum_x_shift .gt. maximum_y_shift) then
            maximum_shift = maximum_x_shift
        else
            maximum_shift = maximum_y_shift
        endif


       if (verbose_output%value) then

            write(*, '(/a,f0.2,a/)')'Finished intial alignment. Maximum shift: ', maximum_shift * binned_pixel_size, ' A'

            write(*,'(a)') 'Refining frame alignment...'

       endif

        total_x_shifts = current_x_shifts
        total_y_shifts = current_y_shifts

        call unblur_refine_alignment(   image_stack, &
                                        maximum_number_of_iterations%value, &
                                        bfactor%value / binned_pixel_size**2, &
                                        mask_central_cross%value, &
                                        width_of_vertical_line%value, &
                                        width_of_horizontal_line%value, &
                                        .false.,0.0e0, & ! no real-space masking of images
                                        0.0e0, & ! no inner radius limit for peak search
                                        outer_radius_after_binning, &
                                        1., &
                                        current_x_shifts,current_y_shifts, &
                                        binned_pixel_size, &
                                        indicate_progress=verbose_output%value)

        total_x_shifts = total_x_shifts + current_x_shifts
        total_y_shifts = total_y_shifts + current_y_shifts


        !
        ! Now let's go back to the intial pixel size and polish off the results
        !
        if (pre_binning_factor .gt. 1) then

            ! The shifts accumulated so far
            total_x_shifts = total_x_shifts * pre_binning_factor
            total_y_shifts = total_y_shifts * pre_binning_factor

            ! Read-in the micrograph frames again and don't resample them, this time

            if (verbose_output%value) then

                write(*,'(a,i0)') 'Finished alignment of frames with binning factor of ', pre_binning_factor
                write(*,'(//a)') 'Will now refine the frame alignment using non-binned images...'

            ! FFT
                write(*,'(a)') 'Applying phase shifts to original, unbinned images...'
                call my_progress_bar%Begin(number_of_frames_per_movie%value)
            endif

            iterations_performed = 0
            !$omp parallel default(shared) private(image_counter)
            !$omp do
            do image_counter = 1,number_of_frames_per_movie%value

                image_stack(image_counter) = unbinned_image_stack(image_counter)
                call image_stack(image_counter)%PhaseShift( total_x_shifts(image_counter), &
                                                            total_y_shifts(image_counter), &
                                                            0e0)
                !$omp atomic
                iterations_performed = iterations_performed + 1
                if (verbose_output%value) call my_progress_bar%Update(iterations_performed)
            enddo

            !$omp enddo
            !$omp end parallel
            if (verbose_output%value) call my_progress_bar%Finish()

            call unblur_refine_alignment(   image_stack, &
                                            maximum_number_of_iterations%value, &
                                            bfactor%value / pixel_size%value**2, &
                                            mask_central_cross%value, &
                                            width_of_vertical_line%value, &
                                            width_of_horizontal_line%value, &
                                            .false.,0.0e0, & ! no real-space masking of images
                                            0.0e0, & ! no inner radius limit for peak search
                                            outer_radius%value / pixel_size%value, &
                                            threshold_maximum_shift%value / pixel_size%value, &
                                            current_x_shifts,current_y_shifts, &
                                            pixel_size%value, &
                                            indicate_progress=verbose_output%value)

            total_x_shifts = total_x_shifts + current_x_shifts
            total_y_shifts = total_y_shifts + current_y_shifts

        else
            ! We are done.
        endif


        ! if necessary, do the drift filter
        if (apply_drift_filter%value) then
            !$omp parallel default(shared) private(image_counter)
            !$omp do
            do image_counter=1, number_of_frames_per_movie%value

                !
                ! 140805
                ! TODO : improve this:
                !   we can get a better guess at the filter by taking halves of each of the two drift vectors (one from the previous, one from the next)
                !   and combining the two resulting drift filters
                !

                if (image_counter .eq. 1) then
                    current_drift_x = total_x_shifts(2) - total_x_shifts(1)
                    current_drift_y = total_y_shifts(2) - total_y_shifts(1)
                else if (image_counter .eq. number_of_frames_per_movie%value) then
                    current_drift_x =  total_x_shifts(number_of_frames_per_movie%value    ) &
                                     - total_x_shifts(number_of_frames_per_movie%value - 1)
                    current_drift_y =  total_y_shifts(number_of_frames_per_movie%value    ) &
                                     - total_y_shifts(number_of_frames_per_movie%value - 1)
                else
                    current_drift_x = 0.5e0 * (total_x_shifts(image_counter + 1) - total_x_shifts(image_counter - 1))
                    current_drift_y = 0.5e0 * (total_y_shifts(image_counter + 1) - total_y_shifts(image_counter - 1))
                endif

                call image_stack(image_counter)%ApplyDriftFilter(current_drift_x, current_drift_y, 0.)
            enddo

            !$omp enddo
            !$omp end parallel
        endif



        ! Dose filtering
        if (apply_dose_filter%value) then
            do image_counter = 1,number_of_frames_per_movie%value
                call my_electron_dose%ApplyDoseFilterToImage(image_stack(image_counter), &
                                                             dose_start=((image_counter-1)*exposure_per_frame%value) + pre_exposure_amount%value, &
                                                             dose_finish=(image_counter*exposure_per_frame%value) + pre_exposure_amount%value, &
                                                             pixel_size=pixel_size%value)
            enddo
        endif



        ! create the final sum and save the aligned images if requird.

        call sum_image%Allocate(mould=image_stack(1),in_real_space=.false.)
        sum_image = (0e0, 0e0)
        sum_image_2 = sum_image


        do image_counter=1, number_of_frames_per_movie%value
            if (IsOdd(image_counter)) then
                call sum_image%AddImage(image_stack(image_counter))
            else
                call sum_image_2%AddImage(image_stack(image_counter))
            endif

            if (save_aligned_frames%value) then
                ! Write out the aligned frames
                !call image_stack(image_counter)%WriteToImageFile(output_aligned_stack_file, (movie_counter-1)*number_of_frames_per_movie%value+image_counter)

                if (movie_counter .eq. 1 .and. image_counter .eq. 1) then
                    call image_stack(image_counter)%WriteToDisk(aligned_frames_filename%value, (movie_counter-1)*number_of_frames_per_movie%value+image_counter, delete_if_already_exists=.true.)
                else
                    call image_stack(image_counter)%WriteToDisk(aligned_frames_filename%value, (movie_counter-1)*number_of_frames_per_movie%value+image_counter)
                endif

            endif

        enddo


        ! FRC
        if (number_of_frames_per_movie%value .gt. 1) then
            fsc_curve = sum_image%GetFSCWith(sum_image_2, 100, pixel_size%value)
            call sum_image%AddImage(sum_image_2)

            call fsc_curve%CopyXData(frc_x_data)
            call fsc_curve%CopyYData(frc_y_data)

            ! Work out the Final score as per original unblur
            final_score_numerator = 0.0e0
            final_score_denominator = 0.0e0
            do bin_counter = 1, 100
                if (frc_x_data(bin_counter) .le. 0.05e0) then
                    final_score_numerator = final_score_numerator + frc_y_data(bin_counter) * real(bin_counter - 1)
                    final_score_denominator = final_score_denominator + real(bin_counter - 1)
                endif
            enddo

            final_score = final_score_numerator / final_score_denominator

            ! write out the fsc_curve if in expert mode

            if (set_expert_options%value) then
                if (movie_counter .eq. 1) then
                    call my_frc_file%Init(frc_output_filename%value, &
                                          OPEN_TO_WRITE, &
                                          fsc_curve%GetNumberOfDataPoints())
                    call my_frc_file%WriteCommentLine('Unblur FRC file for input stack : '//input_filename%value)
                    call my_frc_file%WriteCommentLine('Number of micrographs: '//IntegerToString(number_of_movies))
                    call my_frc_file%WriteCommentLine('Number of frames per movie: '//&
                                                   IntegerToString(number_of_frames_per_movie%value))
                    call my_frc_file%WriteCommentLine('Pixel size (A): '//RealToString(pixel_size%value,4))
                    call my_frc_file%WriteCommentLine('  1/(A)          FRC')
                    call my_frc_file%WriteCommentLine('-----------------------')
                endif

                call my_frc_file%WriteCommentLine('Micrograph '//IntegerToString(movie_counter)//' of '//&
                                              IntegerToString(number_of_movies)//'. Score: '//RealToString(final_score,3))

                call my_frc_file%WriteDataLine(frc_x_data)
                call my_frc_file%WriteDataLine(frc_y_data)
            endif

        else
            final_score = 1.0e0
        endif


       ! If necessary, restore the power

        if (apply_dose_filter%value .and. divide_dose_filter%value) then

            !$omp parallel default(shared) private(j, y, i, x, image_counter, current_denominator, current_critical_dose, current_optimal_dose, current_dose,current_ctf,x_sq,y_sq)
            !$omp do

            ! for this pixel work out what to divide by. first we need the radius

            do j=1, sum_image%GetLogicalDimension(2)
                y = (sum_image%LogicalIndexGivenPhysicalIndexInFourierSpace(j,2) * sum_image%fourier_voxel_size(2))
                y_sq = y**2
                do i=1, sum_image%physical_upper_bound_complex(1)
                    x = ((i-1) * sum_image%fourier_voxel_size(1))
                    x_sq = x**2
                    !
                    if (i .eq. 1 .and. j .eq. 1) then
                        current_critical_dose = critical_dose_at_dc
                    else
                        current_critical_dose = my_electron_dose%CriticalDose(sqrt(x_sq+y_sq) / pixel_size%value)
                    endif


                    current_optimal_dose = my_electron_dose%OptimalDoseGivenCriticalDose(current_critical_dose)

                    ! add up for all frames

                    current_denominator = 0.0e0

                    do image_counter = 1, number_of_frames_per_movie%value

                        current_dose = (real(image_counter) * exposure_per_frame%value) + pre_exposure_amount%value

                        if (apply_dose_filter%value .and. current_dose .lt. current_optimal_dose) then
                            current_denominator = current_denominator &
                                                + my_electron_dose%DoseFilter(current_dose, current_critical_dose)**2
                        endif
!                        if (apply_ctf%value .and. .not. phase_flip_only%value) then
 !                           current_ctf = frame_ctf(image_counter)%EvaluateAtSquaredSpatialFrequency(x_sq+y_sq,atan2(y,x),.false.)
 !                           if (abs(current_ctf) .lt. minimum_absolute_ctf_value) then
 !                               current_ctf = sign(minimum_absolute_ctf_value,current_ctf)
 !                           endif
 !                           current_denominator = current_denominator + current_ctf**2
  !                      endif
                    enddo



                    if (current_denominator .gt. 0.0) then
                        current_denominator = sqrt(current_denominator)
                        sum_image%complex_values(i, j, 1) = sum_image%complex_values(i, j, 1) / current_denominator
                    endif

                enddo
            enddo

            !$omp enddo
            !$omp end parallel

        endif



        ! Write out the sum

        if (movie_counter .eq. 1) then
            call sum_image%WriteToDisk(output_filename%value, movie_counter, delete_if_already_exists=.true.)
        else
            call sum_image%WriteToDisk(output_filename%value, movie_counter)
        endif



        ! Write out shifts file
        if (movie_counter .eq. 1) then
            call my_shifts_file%WriteCommentLine('Unblur shifts file for input stack : '//input_filename%value)
            call my_shifts_file%WriteCommentLine('Shifts below are given in Angstroms')
            call my_shifts_file%WriteCommentLine('Number of micrographs: '//IntegerToString(number_of_movies))
            call my_shifts_file%WriteCommentLine('Number of frames per movie: '//&
                                                  IntegerToString(number_of_frames_per_movie%value))
            call my_shifts_file%WriteCommentLine('Pixel size (A): '//RealToString(pixel_size%value,4))
            call my_shifts_file%WriteCommentLine('2 lines per micrograph. 1: X-Shift (A); 2: Y-Shift (A)')
            call my_shifts_file%WriteCommentLine('-------------------------')
        endif

        call my_shifts_file%WriteCommentLine('Micrograph '//IntegerToString(movie_counter)//&
                                             ' of '//IntegerToString(number_of_movies))
        call my_shifts_file%WriteDataLine(total_x_shifts * pixel_size%value)
        call my_shifts_file%WriteDataLine(total_y_shifts * pixel_size%value)

        ! Output to user
        average_x_shift = sum(abs(total_x_shifts)) / real(size(total_x_shifts))
        average_y_shift = sum(abs(total_y_shifts)) / real(size(total_y_shifts))

    if (.not. verbose_output%value) call my_progress_bar%Update(movie_counter)

    if (verbose_output%value .or. number_of_movies .eq. 1) then
        write(*,'(a,f0.1,a)') 'Average X Shift : ', average_x_shift * pixel_size%value, ' A'
        write(*,'(a,f0.1,a)') 'Average Y Shift : ', average_y_shift * pixel_size%value, ' A'
        write(*,'(a,f0.3  )') 'Final Score     : ', final_score
    endif

    enddo ! end of loop over movies

    if (.not. verbose_output%value) call my_progress_bar%Finish

    ! Deallocation

    if (allocated(image_stack)) deallocate(image_stack)
    if (allocated(current_x_shifts)) deallocate(current_x_shifts)
    if (allocated(current_y_shifts)) deallocate(current_y_shifts)
    if (allocated(total_x_shifts)) deallocate(total_x_shifts)
    if (allocated(total_y_shifts)) deallocate(total_y_shifts)

    call input_imagefile%Close()


    call this_program%Terminate()


end program unblur
